"""Load question answering with reference chains and verbatim.

We believe hallucinations pose a significant issue. Enabling straightforward referencing
of what led to the answer's derivation is imperative.

This chain extracts the information from the documents that was used to answer the
question. The output `source_documents` contains only the documents that were used,
and for each one, only the text fragments that were used to answer are included.
If possible, the list of text fragments that justify the answer is added to
`metadata['verbatims']` for each document.

"""
from .base import QAWithReferencesAndVerbatimsChain
from .retrieval import RetrievalQAWithReferencesAndVerbatimsChain

__all__ = [
    "QAWithReferencesAndVerbatimsChain",
    "RetrievalQAWithReferencesAndVerbatimsChain",
]



# flake8: noqa

from langchain.prompts import PromptTemplate
from langchain.retrievers.multi_query import LineListOutputParser
from .references import references_parser, References

# Use some object. It's easier to update the verbatims schema.
_response_example_1 = References(
    response="This Agreement is governed by English law.",
    documents=[0, 1],
)

_response_example_2 = References(response="", documents=[])

_question_prompt_template = """
Use the following portion of a long document to see if any of the text is relevant to answer the question.
{context}

Question: {question}

Extract all verbatims from texts relevant to answering the question in separate strings else output an empty array.
{format_instructions}

"""
_map_verbatim_parser = LineListOutputParser()

QUESTION_PROMPT = PromptTemplate(
    template=_question_prompt_template,
    input_variables=["context", "question"],
    partial_variables={
        "format_instructions": _map_verbatim_parser.get_format_instructions()
    },
    output_parser=_map_verbatim_parser,
)

_combine_prompt_template = """Given the following extracts from several documents, 
a question and not prior knowledge. 
Process step by step:
- creates a final answer
- produces the json result

QUESTION: Which state/country's law governs the interpretation of the contract?
=========
Content: This Agreement is governed by English law
ids: 0,1
Content: 
ids: 2
Content: 
ids: 3
Content: The english law is applicable for this agreement.
ids:4
=========
FINAL ANSWER:
{response_example_1}

QUESTION: What did the president say about Michael Jackson?
=========
Content: "Madam Speaker, Madam Vice President, our First Lady and Second "
"Gentleman. Members of Congress and the Cabinet."
ids: 0
Content: 
ids: 1,2
Content: 
ids: 3
Content: 
ids: 4
=========
FINAL ANSWER: 
{response_example_2}

QUESTION: {question}
=========
{summaries}
=========
{format_instructions}
FINAL ANSWER:"""
COMBINE_PROMPT = PromptTemplate(
    template=_combine_prompt_template,
    input_variables=["summaries", "question"],
    partial_variables={
        "format_instructions": references_parser.get_format_instructions(),
        "response_example_1": str(_response_example_1),
        "response_example_2": str(_response_example_2),
    },
    output_parser=references_parser,
)

EXAMPLE_PROMPT = PromptTemplate(
    template="Content: {page_content}\n" "Idx: {_idx}",
    input_variables=["page_content", "_idx"],
)



# flake8: noqa
from langchain.prompts import PromptTemplate
from .references import references_parser

_initial_qa_prompt_template = """
Context information is below. 
---------------------
{context_str}
---------------------
Process step by step:
- ignore prior knowledge
- extract references ("IDS")
- create a final answer
- produce the json result

Given the context information the question: {question}

If you don't know the answer, just say that you don't know. 
Don't try to make up an answer.

{format_instructions}
"""

INITIAL_QA_PROMPT = PromptTemplate(
    input_variables=["context_str", "question"],
    template=_initial_qa_prompt_template,
    partial_variables={
        "format_instructions": references_parser.get_format_instructions()
    },
    output_parser=references_parser,
)

_refine_prompt_template = """
Use the following portion from several documents to see if any of the text is relevant to answer the question. 
Given the context information and not prior knowledge answer, the question: {question}

We have provided an existing JSON answer with the list of documents with associated ids: 
{existing_answer}

We have the opportunity to refine the existing answer (only if needed) with some more context below.
------------
{context_str}
------------
Given the new context, refine the original answer to better answer the question. 
If you don't know how to refine the original answer, does not modify the answer.

Process step by step:
- ignore prior knowledge
- with the new context extract references of the new context ("IDS")
- refine the original answer to better answer the question. ONLY if you do update it, append the new IDS and verbatims of texts from the existing answser IDS as well. 
- create a final answer
- produce the json result

ALWAYS return a "IDS" part in your answer. 
If the context isn't useful, return the original answer.

If you don't know the answer, just say that you don't know. Don't try to make up an answer.

{format_instructions}
"""

REFINE_PROMPT = PromptTemplate(
    input_variables=["question", "existing_answer", "context_str"],
    template=_refine_prompt_template,
    partial_variables={
        "format_instructions": references_parser.get_format_instructions()
    },
    output_parser=references_parser,
)


EXAMPLE_PROMPT = PromptTemplate(
    template="ids: {_idx}\n" "Content: {page_content}\n",
    input_variables=["page_content", "_idx"],
)



"""**Chains** are easily reusable components linked together.

Chains encode a sequence of calls to components like models, document retrievers,
other Chains, etc., and provide a simple interface to this sequence.

The Chain interface makes it easy to create apps that are:

    - **Stateful:** add Memory to any Chain to give it state,
    - **Observable:** pass Callbacks to a Chain to execute additional functionality,
      like logging, outside the main sequence of component calls,
    - **Composable:** combine Chains with other components, including other Chains.

**Class hierarchy:**

.. code-block::

    Chain --> <name>Chain  # Examples: LLMChain, MapReduceChain, RouterChain
"""

from .qa_with_references import QAWithReferencesChain, RetrievalQAWithReferencesChain
from .qa_with_references_and_verbatims import (
    QAWithReferencesAndVerbatimsChain,
    RetrievalQAWithReferencesAndVerbatimsChain,
)

__all__ = [
    "QAWithReferencesChain",
    "RetrievalQAWithReferencesChain",
    "QAWithReferencesAndVerbatimsChain",
    "RetrievalQAWithReferencesAndVerbatimsChain",
]



# flake8: noqa

from langchain.output_parsers import PydanticOutputParser
from langchain.prompts import PromptTemplate
from langchain.schema import BaseOutputParser
from .verbatims import VerbatimsFromDoc, verbatims_parser, Verbatims

_map_verbatim_parser: BaseOutputParser = PydanticOutputParser(
    pydantic_object=VerbatimsFromDoc
)

# Use some object. It's easier to update the verbatims schema.
_response_example_1 = Verbatims(
    response="This Agreement is governed by English law.",
    documents=[
        VerbatimsFromDoc(
            ids=[0],
            verbatims=[
                "This Agreement is governed by English law",
                # "The english law is applicable for this agreement."
            ],
        ),
        VerbatimsFromDoc(
            ids=[4], verbatims=["The english law is applicable for this agreement."]
        ),
    ],
)

_response_example_2 = Verbatims(response="", documents=[])

_sample_1 = VerbatimsFromDoc(
    ids=[0, 1], verbatims=["This Agreement is governed by English law"]
)
_sample_2 = VerbatimsFromDoc(ids=[2], verbatims=[])
_sample_3 = VerbatimsFromDoc(ids=[3], verbatims=[])
_sample_4 = VerbatimsFromDoc(
    ids=[4], verbatims=["The english law is applicable for this agreement."]
)

_sample_5 = VerbatimsFromDoc(
    ids=[0],
    verbatims=[
        "Madam Speaker, Madam Vice President, our First Lady and Second "
        "Gentleman. Members of Congress and the Cabinet."
    ],
)
_sample_6 = VerbatimsFromDoc(ids=[1, 2], verbatims=[])
_sample_7 = VerbatimsFromDoc(ids=[3], verbatims=[])
_sample_8 = VerbatimsFromDoc(ids=[4], verbatims=[])

_question_prompt_template = """
Use the following portion of a long document to see if any of the text is relevant to answer the question.
{context}

Question: {question}

Extract all verbatims from texts relevant to answering the question in a list of strings  else output an empty array.
{format_instructions}

"""

QUESTION_PROMPT = PromptTemplate(
    template=_question_prompt_template,
    input_variables=["context", "question"],
    partial_variables={
        "format_instructions": _map_verbatim_parser.get_format_instructions()
    },
    output_parser=_map_verbatim_parser,
)

_combine_prompt_template = """Given the following extracts from several documents, 
a question and not prior knowledge. 
Process step by step:
- extract all the verbatims from the texts only if they are relevant to answering the question, in a list of strings  
- creates a final answer
- produces the json result

QUESTION: Which state/country's law governs the interpretation of the contract?
=========
Content: {_sample_1}
Content: {_sample_2}
Content: {_sample_3}
Content: {_sample_4}
=========
FINAL ANSWER:
{response_example_1}

QUESTION: What did the president say about Michael Jackson?
=========
Content: {_sample_5}
Content: {_sample_6}
Content: {_sample_7}
Content: {_sample_8}
=========
FINAL ANSWER: 
{response_example_2}

QUESTION: {question}
=========
{summaries}
=========
{format_instructions}
FINAL ANSWER:"""
COMBINE_PROMPT = PromptTemplate(
    template=_combine_prompt_template,
    input_variables=["summaries", "question"],
    partial_variables={
        "format_instructions": verbatims_parser.get_format_instructions(),
        "response_example_1": _response_example_1.json(),
        "response_example_2": _response_example_2.json(),
        "_sample_1": _sample_1.json(),
        "_sample_2": _sample_2.json(),
        "_sample_3": _sample_3.json(),
        "_sample_4": _sample_4.json(),
        "_sample_5": _sample_5.json(),
        "_sample_6": _sample_6.json(),
        "_sample_7": _sample_7.json(),
        "_sample_8": _sample_8.json(),
    },
    output_parser=verbatims_parser,
)

EXAMPLE_PROMPT = PromptTemplate(
    template="Content: {page_content}\n" "Idx: {_idx}",
    input_variables=["page_content", "_idx"],
)



"""Load question answering with reference chains.

We believe hallucinations pose a significant issue. Enabling straightforward referencing
of what led to the answer's derivation is imperative.

This chain extracts the information from the documents that was used to answer the
question. The output `source_documents` contains only the documents that were used.
"""
from .base import QAWithReferencesChain
from .retrieval import RetrievalQAWithReferencesChain

__all__ = ["QAWithReferencesChain", "RetrievalQAWithReferencesChain"]



# flake8: noqa
from langchain.output_parsers import PydanticOutputParser, RegexParser
from langchain.prompts import PromptTemplate
from langchain.schema import BaseOutputParser
from .verbatims import VerbatimsFromDoc, Verbatims

_map_verbatim_parser: BaseOutputParser = PydanticOutputParser(
    pydantic_object=VerbatimsFromDoc
)

_rank_parser = RegexParser(
    regex=r"(.*?)\nScore: (\d*)",
    output_keys=["answer", "score"],
)


_response_example_1 = Verbatims(
    response="red", documents=[VerbatimsFromDoc(ids=[99], verbatims=["Apples are red"])]
)
_response_example_2 = Verbatims(
    response="a sports car or an suv",
    documents=[
        VerbatimsFromDoc(
            ids=[99], verbatims=["he was not sure if it was a sports car or an suv"]
        )
    ],
)
_response_example_3 = Verbatims(
    response="This document does not answer the question", documents=[]
)
prompt_template = """
Given the following extracts from several documents, a question and not prior knowledge. 

How to determine the score:
- Higher is a better answer
- Better responds fully to the asked question, with sufficient level of detail
- If you do not know the answer based on the context, that should be a score of 0
- Don't be overconfident!

Process step by step:
- extract the references ("IDS")
- extract all the verbatims from the texts only if they are relevant to answering the question, in a list of strings 
- answers the question
- calculates a score of how fully it answered the user's question
- creates a final answer


Here the format for the Helpful Answer.
{format_instructions}

This should be in the following format:
Question: [question here]
Helpful Answer: [json answer here]
Score: [score between 0 and 100]

Example #1
Context:
---------
Apples are red. The car is blue.
---------
Question: what color are apples?
Helpful Answer: {response_example_1}
Score: 100

Example #2
Context:
---------
it was night and the witness forgot his glasses. he was not sure if it was a sports car or an suv
---------
Question: what type was the car?
Helpful Answer: {response_example_2}
Score: 60

Example #3
Context:
---------
Pears are either red or orange
---------
Question: what color are apples?
Helpful Answer: {response_example_3}
Score: 0

Begin!

Context:
---------
{context}
---------
Question: {question}
Helpful Answer:"""
PROMPT = PromptTemplate(
    template=prompt_template,
    input_variables=["context", "question"],
    partial_variables={
        "format_instructions": _map_verbatim_parser.get_format_instructions(),
        "response_example_1": _response_example_1.json(),
        "response_example_2": _response_example_2.json(),
        "response_example_3": _response_example_3.json(),
    },
    output_parser=_rank_parser,
)

EXAMPLE_PROMPT = PromptTemplate(
    template="Content: {page_content}\n" "Idx: {_idx}",
    input_variables=["page_content", "_idx"],
)



"""Question answering with references over documents."""

from __future__ import annotations

import inspect
import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Set, Tuple, cast

from langchain.base_language import BaseLanguageModel
from langchain.callbacks.manager import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
)
from langchain.chains.base import Chain
from langchain.chains.combine_documents.base import BaseCombineDocumentsChain
from langchain.chains.combine_documents.map_reduce import MapReduceDocumentsChain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.chains.llm import LLMChain
from langchain.docstore.document import Document

# Impossible to import in experimental. bug in the CI
from langchain.pydantic_v1 import Extra

# from pydantic import Extra
from langchain.schema import BaseOutputParser, BasePromptTemplate, OutputParserException

from .loading import (
    load_qa_with_references_chain,
)
from .map_reduce_prompts import (
    COMBINE_PROMPT,
    EXAMPLE_PROMPT,
    QUESTION_PROMPT,
)
from .references import references_parser

logger = logging.getLogger(__name__)


# TOTRY: Add in VectorStoreIndexWrapper.query_with_references()
# TOTRY: Add in VectorStoreIndexWrapper.query_with_references_and_verbatims()
class BaseQAWithReferencesChain(Chain, ABC):
    combine_documents_chain: BaseCombineDocumentsChain

    """Chain to use to combine documents."""
    question_key: str = "question"  #: :meta private:
    input_docs_key: str = "docs"  #: :meta private:
    answer_key: str = "answer"  #: :meta private:
    source_documents_key: str = "source_documents"  #: :meta private:
    chain_type: str  #: :meta private:
    output_parser: Optional[BaseOutputParser] = None

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        document_prompt: BasePromptTemplate = EXAMPLE_PROMPT,
        question_prompt: BasePromptTemplate = QUESTION_PROMPT,
        combine_prompt: BasePromptTemplate = COMBINE_PROMPT,
        **kwargs: Any,
    ) -> BaseQAWithReferencesChain:
        """Construct the chain from an LLM."""
        llm_question_chain = LLMChain(llm=llm, prompt=question_prompt)
        llm_combine_chain = LLMChain(llm=llm, prompt=combine_prompt)
        combine_results_chain = StuffDocumentsChain(
            llm_chain=llm_combine_chain,
            document_prompt=document_prompt,
            document_variable_name="summaries",
        )
        combine_document_chain = MapReduceDocumentsChain(
            llm_chain=llm_question_chain,
            combine_document_chain=combine_results_chain,
            document_variable_name="context",
            return_intermediate_steps=True,
        )
        return cls(
            combine_documents_chain=combine_document_chain,
            **kwargs,
        )

    @classmethod
    def from_chain_type(
        cls,
        llm: BaseLanguageModel,
        chain_type: str = "stuff",
        chain_type_kwargs: Optional[dict] = None,
        **kwargs: Any,
    ) -> BaseQAWithReferencesChain:
        """Load chain from chain type."""
        _chain_kwargs = chain_type_kwargs or {}
        combine_document_chain = load_qa_with_references_chain(
            llm,
            chain_type=chain_type,
            **_chain_kwargs,
        )
        if chain_type == "map_rerank" and "output_parser" not in kwargs:
            from .map_rerank_prompts import rerank_reference_parser

            kwargs["output_parser"] = rerank_reference_parser

        return cls(
            combine_documents_chain=combine_document_chain,
            chain_type=chain_type,
            **kwargs,
        )

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    @property
    def input_keys(self) -> List[str]:
        """Expect input key.

        :meta private:
        """
        return [self.question_key]

    @property
    def output_keys(self) -> List[str]:
        """Return output key.

        :meta private:
        """
        _output_keys = [self.answer_key, self.source_documents_key]
        return _output_keys

    @abstractmethod
    def _get_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: CallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""

    def _process_reference(
        self, answers: Dict[str, Any], docs: List[Document], references: Any
    ) -> Set[int]:
        # With the map_rerank mode, use extra parameter for map_rerank to identify
        # the corresponding document.
        if "_idx" in answers:
            references.documents.append(answers["_idx"])

        return references.documents

    def _process_results(
        self,
        answers: Dict[str, Any],
        docs: List[Document],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Tuple[str, Set[int]]:
        idx = set()
        answer = answers[self.combine_documents_chain.output_key]
        try:
            # At this time (version 0.0.273), the output_parser is not
            # automatically called.
            # We must extract this parser to analyse the response.
            parser: Optional[BaseOutputParser] = self.output_parser
            if not parser:
                if self.chain_type == "map_rerank":
                    assert self.output_parser
                elif self.chain_type == "map_reduce":
                    parser = cast(
                        Any, self.combine_documents_chain
                    ).collapse_document_chain.llm_chain.prompt.output_parser
                elif self.chain_type == "refine":
                    parser = cast(
                        Any, self.combine_documents_chain
                    ).refine_llm_chain.prompt.output_parser
                else:
                    # Simple chain
                    parser = cast(
                        Any, self.combine_documents_chain
                    ).llm_chain.prompt.output_parser
            assert parser
            references = parser.parse(answer)
            answer = references.response

            idx = set(self._process_reference(answers, docs, references))

            # Purge _idx
            for doc in docs:
                del doc.metadata["_idx"]
            return answer, idx
        except OutputParserException as e:
            # Probably that the answer has been cut off.
            if run_manager:
                run_manager.on_chain_error(e)
            raise e
        except Exception as e:
            if run_manager:
                run_manager.on_chain_error(e)
            raise e

    @abstractmethod
    async def _aget_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: AsyncCallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        accepts_run_manager = (
            "run_manager" in inspect.signature(self._get_docs).parameters
        )
        if accepts_run_manager:
            docs = self._get_docs(inputs, run_manager=_run_manager)
        else:
            docs = self._get_docs(inputs)  # type: ignore[call-arg]

        # Inject position in the list
        for idx, doc in enumerate(docs):
            doc.metadata["_idx"] = idx

        answers = self.combine_documents_chain(
            {
                self.combine_documents_chain.input_key: docs,
                self.question_key: inputs[self.question_key],
            },
            callbacks=_run_manager.get_child(),
        )
        answer, all_idx = self._process_results(answers, docs)
        selected_docs = [docs[idx] for idx in all_idx if idx < len(docs)]

        return {
            self.answer_key: answer,
            self.source_documents_key: selected_docs,
        }

    async def _acall(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        _run_manager = run_manager or AsyncCallbackManagerForChainRun.get_noop_manager()
        accepts_run_manager = (
            "run_manager" in inspect.signature(self._aget_docs).parameters
        )
        if accepts_run_manager:
            docs = await self._aget_docs(inputs, run_manager=_run_manager)
        else:
            docs = await self._aget_docs(inputs)  # type: ignore[call-arg]

        # Inject position in the list
        for idx, doc in enumerate(docs):
            doc.metadata["_idx"] = idx

        answers = await self.combine_documents_chain.acall(
            {
                self.combine_documents_chain.input_key: docs,
                self.input_keys[0]: inputs[self.input_keys[0]],
            },
            callbacks=_run_manager.get_child(),
        )
        answer, all_idx = self._process_results(answers, docs)
        selected_docs = [docs[idx] for idx in all_idx if idx < len(docs)]

        return {
            self.answer_key: answer,
            self.source_documents_key: selected_docs,
        }


class QAWithReferencesChain(BaseQAWithReferencesChain):
    """
    This chain extracts the information from the documents that was used to answer the
    question. The output `source_documents` contains only the documents that were used,
    and for each one.

    The result["source_documents"] returns only the list of documents used.
    Then it is possible to find the page of a PDF, or the chapter of a markdown.
    """

    input_docs_key: str = "docs"  #: :meta private:

    @property
    def input_keys(self) -> List[str]:
        """Expect input key.

        :meta private:
        """
        return [self.input_docs_key, self.question_key]

    def _get_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: CallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""
        return inputs.pop(self.input_docs_key)

    async def _aget_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: AsyncCallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""
        return inputs.pop(self.input_docs_key)

    @property
    def _chain_type(self) -> str:
        return "qa_with_references_chain"



"""Question-answering with sources over an index."""

from ..qa_with_references.retrieval import RetrievalQAWithReferencesChain
from .base import (
    BaseQAWithReferencesAndVerbatimsChain,
)


class RetrievalQAWithReferencesAndVerbatimsChain(
    RetrievalQAWithReferencesChain, BaseQAWithReferencesAndVerbatimsChain
):
    """Question-answering with referenced documents and verbatim.

    This implementation allows you to retrieve the list of documents, enriched with
    the verbatim metadata.
    The implementation uses fewer tokens and correctly handles recursive map_reduces.
    """



# flake8: noqa


from langchain.prompts import PromptTemplate
from .references import references_parser, References

_response_example_1 = References(
    response="This Agreement is governed by English law.",
    documents=[0],
)

# Sample with no result
_response_example_2 = References(response="I don't know.", documents=[])

_template = """Given the following extracts from several documents, a question and not prior knowledge. 
Process step by step:
- for each documents extract the references ("IDS")
- creates a final answer
- produces the json result

If you don't know the answer, just say that you don't know. Don't try to make up an answer.
ALWAYS return a "IDS" part in your answer.

QUESTION: Which state/country's law governs the interpretation of the contract?
=========
Content: This Agreement is governed by English law and the parties submit to the exclusive jurisdiction of the English courts in relation to any dispute (contractual or non-contractual) concerning this Agreement save that either party may apply to any court for an injunction or other relief to protect its Intellectual Property Rights.
Ids: 0

Content: No Waiver. Failure or delay in exercising any right or remedy under this Agreement shall not constitute a waiver of such (or any other) right or remedy.
Ids: 1

Content: (b) if Google believes, in good faith, that the Distributor has violated or caused Google to violate any Anti-Bribery Laws
Ids: 2

Content: The english law is applicable for this agreement.
Ids: 3
=========
FINAL ANSWER: 
```
{_response_example_1}
```

QUESTION: What did the president say about Michael Jackson?
=========
Content: Madam Speaker, Madam Vice President, our First Lady and Second Gentleman. Members of Congress and the Cabinet.
Ids: 0

Content: And we won’t stop. \n\nWe have lost so much to COVID-19. Time with one another. And worst of all, so much loss of life.
Ids: 1

Content: And a proud Ukrainian people, who have known 30 years of independence, have repeatedly shown that they will not tolerate anyone who tries to take their country backwards.
Ids: 2

Content: More support for patients and families. \n\nTo get there, I call on Congress to fund ARPA-H, the Advanced Research Projects Agency for Health.
ids: 3
=========
FINAL ANSWER: 
```
{_response_example_2}
```

QUESTION: {question}
=========
{summaries}
=========
{format_instructions}
FINAL ANSWER: 
"""

PROMPT = PromptTemplate(
    template=_template,
    input_variables=["summaries", "question"],
    partial_variables={
        "format_instructions": references_parser.get_format_instructions(),
        "_response_example_1": str(_response_example_1),
        "_response_example_2": str(_response_example_2),
    },
    output_parser=references_parser,
)
EXAMPLE_PROMPT = PromptTemplate(
    template="Content: {page_content}\n" "idx: {_idx}",
    input_variables=["page_content", "_idx"],
)



"""Question answering with references and verbatims over documents."""

from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Set, cast

from langchain.base_language import BaseLanguageModel
from langchain.callbacks.manager import (
    AsyncCallbackManagerForChainRun,
    CallbackManagerForChainRun,
)
from langchain.chains.combine_documents.map_reduce import MapReduceDocumentsChain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.chains.llm import LLMChain
from langchain.docstore.document import Document
from langchain.schema import BasePromptTemplate, OutputParserException

from ..qa_with_references.base import BaseQAWithReferencesChain
from .loading import (
    load_qa_with_references_chain,
)
from .map_reduce_prompts import (
    COMBINE_PROMPT,
    EXAMPLE_PROMPT,
    QUESTION_PROMPT,
)
from .verbatims import Verbatims, verbatims_parser

logger = logging.getLogger(__name__)


class BaseQAWithReferencesAndVerbatimsChain(BaseQAWithReferencesChain):
    def _process_reference(
        self, answers: Dict[str, Any], docs: List[Document], references: Any
    ) -> Set[int]:
        idx = set()
        try:
            verbatims = cast(Verbatims, references)
            # With the map_rerank mode, use extra parameter for map_rerank to identify
            # the corresponding document.
            if "_idx" in answers:
                # Use extra parameter for map_rerank to identify the corresponding document.
                # Fix the ids of the selected document.
                verbatims.documents[0].ids = [answers["_idx"]]

            # With the map_reduce mode, we must inject the associated idx.
            if self.chain_type == "map_reduce":
                parser_doc = cast(
                    Any, self.combine_documents_chain
                ).llm_chain.prompt.output_parser
                for i, str_verbatim in enumerate(answers["intermediate_steps"]):
                    doc_ref = parser_doc.parse(str_verbatim)
                    for d in verbatims.documents:
                        if d.verbatims == doc_ref.verbatims:
                            d.ids = [docs[i].metadata["_idx"]]

            # Inject verbatims and get idx
            for ref_doc in verbatims.documents:
                for doc_id in ref_doc.ids:
                    if 0 <= doc_id < len(docs):  # Guard
                        if ref_doc.verbatims:
                            idx.add(doc_id)
                            # Search the real verbatim if possible
                            original_verbatim = ref_doc.original_verbatims(
                                docs[doc_id].page_content
                            )
                            if original_verbatim:
                                docs[doc_id].metadata["verbatims"] = original_verbatim
            return idx
        except OutputParserException as e:
            logger.debug(f"Exception during parsing: {e}")
            return idx

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        document_prompt: BasePromptTemplate = EXAMPLE_PROMPT,
        question_prompt: BasePromptTemplate = QUESTION_PROMPT,
        combine_prompt: BasePromptTemplate = COMBINE_PROMPT,
        **kwargs: Any,
    ) -> BaseQAWithReferencesChain:
        """Construct the chain from an LLM."""
        llm_question_chain = LLMChain(llm=llm, prompt=question_prompt)
        llm_combine_chain = LLMChain(llm=llm, prompt=combine_prompt)
        combine_results_chain = StuffDocumentsChain(
            llm_chain=llm_combine_chain,
            document_prompt=document_prompt,
            document_variable_name="summaries",
        )
        combine_document_chain = MapReduceDocumentsChain(
            llm_chain=llm_question_chain,
            combine_document_chain=combine_results_chain,
            document_variable_name="context",
            return_intermediate_steps=True,
        )
        return cls(
            combine_documents_chain=combine_document_chain,
            **kwargs,
        )

    @classmethod
    def from_chain_type(
        cls,
        llm: BaseLanguageModel,
        chain_type: str = "stuff",
        chain_type_kwargs: Optional[dict] = None,
        **kwargs: Any,
    ) -> BaseQAWithReferencesChain:
        """Load chain from chain type."""
        _chain_kwargs = chain_type_kwargs or {}
        combine_document_chain = load_qa_with_references_chain(
            llm,
            chain_type=chain_type,
            **_chain_kwargs,
        )
        if chain_type == "map_rerank" and "output_parser" not in kwargs:
            kwargs["output_parser"] = verbatims_parser

        return cls(
            combine_documents_chain=combine_document_chain,
            chain_type=chain_type,
            **kwargs,
        )


class QAWithReferencesAndVerbatimsChain(BaseQAWithReferencesAndVerbatimsChain):
    """
    Question answering with references and verbatims over documents.

    This chain extracts the information from the documents that was used to answer the
    question. The output `source_documents` contains only the documents that were used,
    and for each one, only the text fragments that were used to answer are included.
    If possible, the list of text fragments that justify the answer is added to
    `metadata['verbatims']` for each document.
    Then it is possible to find the page of a PDF, or the chapter of a markdown.

    The incogenent of this chain, and that this consumes output tokens.

    A sample result of usage with Wikipedia, may be:
    ```
    For the question "what can you say about ukraine?",
    to answer "Ukraine has an illegal and unprovoked invasion inside its territory,
    and its citizens show courage and are fighting against it. It is believed that
    the capital city of Kyiv, which is home to 2.8 million people, is a target.",
    the LLM use:
    Source https://www.defense.gov/News/Transcripts/...
    -  "when it comes to conducting their illegal and unprovoked invasion inside
    Ukraine."
    Source https://www.whitehouse.gov/briefing-room/...
    - "to the fearless and skilled Ukrainian fighters who are standing in the
    breach"
    - "You got to admit, you have -- must be amazed at the courage of this country"
    Source https://www.whitehouse.gov/briefing-room/...
    -  "believe that they will target Ukraine's capital, Kyiv, a city of 2.8 million
    innocent people."
    Source https://www.whitehouse.gov/briefing-room/...
    -  "believe that they will target Ukraine's capital, Kyiv, a city of 2.8 million
    innocent people."
    ```
    """

    input_docs_key: str = "docs"  #: :meta private:

    @property
    def input_keys(self) -> List[str]:
        """Expect input key.

        :meta private:
        """
        return [self.input_docs_key, self.question_key]

    def _get_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: CallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""
        return inputs.pop(self.input_docs_key)

    async def _aget_docs(
        self,
        inputs: Dict[str, Any],
        *,
        run_manager: AsyncCallbackManagerForChainRun,
    ) -> List[Document]:
        """Get docs to run questioning over."""
        return inputs.pop(self.input_docs_key)

    @property
    def _chain_type(self) -> str:
        return "qa_with_references_chain"



import logging
from typing import List

from langchain.output_parsers import PydanticOutputParser, RegexParser

# Impossible to import in experimental. bug in the CI
from langchain.pydantic_v1 import BaseModel

# from pydantic import BaseModel
from langchain.schema import BaseOutputParser
from pydantic import Field

logger = logging.getLogger(__name__)

# To optimize the consumption of tokens, it's better to use only 'text', without json.
# Else the schema consume ~300 tokens and the response 20 tokens by step
_OPTIMIZE = True


class References(BaseModel):
    """
    Response and referenced documents.
    """

    response: str
    """ The response """
    documents: List[int] = []
    """ The list of documents used to response """

    def __str__(self) -> str:
        if _OPTIMIZE:
            return f'{self.response}\nIDS:{",".join(map(str,self.documents))}'
        else:
            return self.json()


references_parser: BaseOutputParser
if _OPTIMIZE:

    class _ReferencesParser(RegexParser):
        """An optimised parser for Reference.
        It's more effective than the pydantic approach
        """

        def get_format_instructions(self) -> str:
            return ""

        def parse(self, text: str) -> References:
            dict = super().parse(text)
            if dict["ids"]:
                ids = [int(d) for d in map(str, dict["ids"].split(","))]
            else:
                ids = []
            return References(response=dict["response"], documents=ids)

    references_parser = _ReferencesParser(
        regex=r"(.*)\nIDS:(.*)",
        output_keys=["response", "ids"],
        default_output_key="response",
    )
else:
    references_parser = PydanticOutputParser(pydantic_object=References)



import logging
import re
from typing import List, Optional

from langchain.output_parsers import PydanticOutputParser

# Impossible to import in experimental. bug in the CI
from langchain.pydantic_v1 import BaseModel

# from pydantic import BaseModel
from langchain.schema import BaseOutputParser

logger = logging.getLogger(__name__)


def _extract_original_verbatim(verbatim: str, page_content: str) -> Optional[str]:
    """The exact format of verbatim may be changed by the LLM.
    Extract only the words of the verbatim, and try to find a sequence
    of same words in the original document.
    """
    only_words = filter(len, re.split(r"[^\w]+", verbatim))
    regex_for_words_in_same_oder = (
        r"(?i)\b" + r"\b[^\w]+".join(only_words) + r"\b"
        r"\s*[.!?:;]?"
        # Optional end of sentence
    )
    match = re.search(regex_for_words_in_same_oder, page_content, re.IGNORECASE)
    if match:
        return match[0].strip()
    return None  # No verbatim found in the original document


class VerbatimsFromDoc(BaseModel):
    ids: List[int]
    """ The position of the document in the original list """
    verbatims: List[str]
    """ All verbatims for this document """

    def original_verbatims(self, page_content: str) -> List[str]:
        result = []
        for j, verbatim in enumerate(self.verbatims):
            original_verbatim = _extract_original_verbatim(
                verbatim=verbatim, page_content=page_content
            )
            if original_verbatim:
                result.append(original_verbatim)
            else:
                logger.debug(f'Ignore verbatim "{verbatim}"')
        return result

    def __str__(self) -> str:
        return "\n".join(
            [f'"{v}"' for v in [re.sub(r"\s", " ", v) for v in self.verbatims]]
        )


class Verbatims(BaseModel):
    """
    Response, references and verbatims object.
    """

    response: str
    """ The response """
    documents: List[VerbatimsFromDoc]
    """ The list of documents and verbatims"""


verbatims_parser: BaseOutputParser = PydanticOutputParser(pydantic_object=Verbatims)
""" A parser for this object """



# flake8: noqa
from typing import Type

from langchain.output_parsers import RegexParser
from langchain.prompts import PromptTemplate
from langchain.schema import BaseOutputParser
from .references import References

_rank_parser = RegexParser(
    regex=r"(.*?)\nScore: (\d*)",
    output_keys=["answer", "score"],
)


class ReferenceOutputParser(BaseOutputParser[References]):
    """Parse an output using a pydantic model."""

    pydantic_object: Type[References] = References
    """The pydantic model to parse."""

    def parse(self, text: str) -> References:
        return References(response=text)

    def get_format_instructions(self) -> str:
        return ""

    @property
    def _type(self) -> str:
        return "reference"


rerank_reference_parser = ReferenceOutputParser()

prompt_template = """
Given the following extracts from several documents, a question and not prior knowledge. 

How to determine the score:
- Higher is a better answer
- Better responds fully to the asked question, with sufficient level of detail
- If you do not know the answer based on the context, that should be a score of 0
- Don't be overconfident!

Process step by step:
- answers the question
- calculates a score of how fully it answered the user's question
- creates a final answer


This should be in the following format:
Question: [question here]
Helpful Answer: [json answer here]
Score: [score between 0 and 100]

Example #1
Context:
---------
Apples are red. The car is blue.
---------
Question: what color are apples?
Helpful Answer: red
Score: 100

Example #2
Context:
---------
it was night and the witness forgot his glasses. he was not sure if it was a sports car or an suv
---------
Question: what type was the car?
Helpful Answer: a sports car or an suv
Score: 60

Example #3
Context:
---------
Pears are either red or orange
---------
Question: what color are apples?
Helpful Answer: This document does not answer the question
Score: 0

Begin!

Context:
---------
{context}
---------
Question: {question}
Helpful Answer:"""
PROMPT = PromptTemplate(
    template=prompt_template,
    input_variables=["context", "question"],
    output_parser=_rank_parser,
)

EXAMPLE_PROMPT = PromptTemplate(
    template="Content: {page_content}\nIdx: {_idx}\n",
    input_variables=["page_content", "_idx"],
)
